use std::fs;

const GRID_WIDTH: i32 = 1000;
const GRID_HEIGHT: i32 = 1000;
const START_X: i32 = 500;
const START_Y: i32 = 500;


pub fn solve() {
    let path = "res/09/input.txt";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");

    let lines: Vec<&str> = contents
        .lines()
        .filter(|line| *line != "")
        .collect();

    // task 1 - rope with 2 knots
    let mut rope2 = Rope::new(2);

    // task 2 - rope with 10 knots
    let mut rope10 = Rope::new(10);

    lines.iter().for_each(|line| {
        // parse command
        let cmd: Vec<&str> = line.split_whitespace().collect();
        let dir = match cmd[0] {
            "U" => Dir::Up,
            "D" => Dir::Down,
            "L" => Dir::Left,
            "R" => Dir::Right,
            _ => Dir::Up
        };
        let amount = cmd[1].parse::<i32>().unwrap();

        // move head
        for _ in 0..amount {
            rope2.mv(dir);
            rope10.mv(dir);
        }
    });

    // task 1 res
    let res = rope2.tail_visited_num();
    println!("Result 1: {}", res);

    // task 2 res
    let res = rope10.tail_visited_num();
    println!("Result 1: {}", res);



}

#[derive(Debug, Copy, Clone)]
enum Dir {
    Up,
    Down,
    Left,
    Right
}

struct Rope {
    knots_x: Vec<i32>,
    knots_y: Vec<i32>,
    tail_visited_grid: Vec<Vec<bool>> // [Y][X]
}

impl Rope {
    pub fn new(knots: i32) -> Self {
        Rope {
            knots_x: (0..knots).map(|_| START_X).collect(),
            knots_y: (0..knots).map(|_| START_Y).collect(),
            tail_visited_grid: (0..GRID_HEIGHT)
                .map(|_| (0..GRID_WIDTH)
                    .map(|_| false).collect()
                ).collect()
        }
    }

    // move head (first knot)
    pub fn mv(&mut self, dir: Dir) {
        match dir {
            Dir::Up => {
                self.knots_y[0] -= 1
            },
            Dir::Down => {
                self.knots_y[0] += 1
            },
            Dir::Left => {
                self.knots_x[0] -= 1
            },
            Dir::Right => {
                self.knots_x[0] += 1
            }
        }
        self.check_tail_pos()
    }

    // move tail if necessary
    fn check_tail_pos(&mut self) {
        // check pos of each knot except first
        for i in 1..self.knots_x.len() {
            if (self.knots_x[i-1] == self.knots_x[i] && (self.knots_y[i-1] - self.knots_y[i]).abs() == 1)
                    || (self.knots_y[i-1] == self.knots_y[i] && (self.knots_x[i-1] - self.knots_x[i]).abs() == 1) {
                // Knots are touching, nothing to do
            } else if (self.knots_x[i-1] - self.knots_x[i]).abs() == 1 && (self.knots_y[i-1] - self.knots_y[i]).abs() == 1 {
                // Knots are touching diagonally, nothing to do
            } else {
                // Tail needs to be moved
                if self.knots_x[i-1] != self.knots_x[i] && self.knots_y[i-1] != self.knots_y[i] {
                    // Move diagonally
                    self.knots_x[i] += normalize(self.knots_x[i-1] - self.knots_x[i]);
                    self.knots_y[i] += normalize(self.knots_y[i-1] - self.knots_y[i]);
                } else {
                    // Move vertically/horizontally
                    if self.knots_x[i-1] == self.knots_x[i] {
                        self.knots_y[i] += normalize(self.knots_y[i-1] - self.knots_y[i]);
                    } else {
                        self.knots_x[i] += normalize(self.knots_x[i-1] - self.knots_x[i]);
                    }
                }
            }
        }

        // Set current tail (last knot) pos as visited
        self.tail_visited_grid[*self.knots_y.last().unwrap() as usize][*self.knots_x.last().unwrap() as usize] = true;        
    }

    // returns number of cells visited by tail
    pub fn tail_visited_num(&self) -> i32 {
        self.tail_visited_grid.iter()
            .map(|row| 
                row.iter()
                .map(|cell| 
                    if *cell { 1 as i32 } 
                    else { 0 as i32 }
                ).sum::<i32>()
            ).sum()
    }
}

fn normalize(n: i32) -> i32 {
    if n > 0 {
        1
    } else if n < 0 {
        -1
    } else {
        0
    }
}